/******************************************************************
Yun Zeng, 2011
theory.of.computation@gmail.com

MRF optimization for Triagulated graphs
Energy:
\sum_{i \in V} \theta_i(x_i) + \sum_{(i,j,k)\in F} \theta_{ijk}(x_i, x_j, x_k)
*******************************************************************/
#ifndef __TRIMRFENERGY_H__
#define __TRIMRFENERGY_H__

#include <vector>
#include <iostream>

#define CUDA_OPTIMIZE 1 // uncomment this line if no hardware support

#ifdef CUDA_OPTIMIZE
extern "C" void cudaLoadEnergy (int V, int F, int L, float *theta_single, float *theta_tri, int *fIdxArray) ;
extern "C" void cudaGetEnergy (int V, int F, int L, float *theta_single, float *theta_tri);
extern "C" void cudaOptimizeTriMRF (int V, int F, int L, float eps, int maxItr);
extern "C" void cudaOptimizeTriMRFNew (int V, int F, int L, float eps, int maxItr, std::vector<std::vector<int> > &ff_list);
extern "C" void cudaClearTriMRF ();
extern "C" void cudaInitTimer () ;
extern "C" void cudaBeginTimer ();
extern "C" void cudaEndTimer ();
extern "C" void cudaDestroyTimer ();
#endif

class TriMRFEnergy
{
public: // Variables and Data structure
	enum OPTIMIZER {CPU_RANDOM_ORDER, CPU_FIXED_ORDER, GPU, CPU_RANDOM_PERFACE};
	struct Options
	{
		Options()
		{	// default parameters
			m_eps = 0.001f; 
			m_iterMax = 3000;
			m_printIter = 10;
			optimizer = TriMRFEnergy::CPU_RANDOM_ORDER;
		}

		// stopping criterion
		float	m_eps; // stop if the increase in the lower bound during one iteration is less or equal than m_eps.						   
		int		m_iterMax; // maximum number of iterations
		int		m_printIter; // print lower bound and energy every m_printIter iterations
		OPTIMIZER optimizer;
	};

public: // public functions
	// constructor.
	TriMRFEnergy(int v_num, int f_num, int l_num);
	// destructor.
	~TriMRFEnergy();

	// add node idx for face fID: (vID1, vID2, vID3)
	void AddFaceInd2Nd(int fID, int vID1, int vID2, int vID3);

	// add singleton cost for node nID
	// size of data: L
	void AddNodeData(int nID, float *data);

	// add triplet cost for face fID
	// Order to store \theta_tri
	//     L0, L1, L2
	// 1	0 0 0
	// 2	1 0 0
	// 3	2 0 0
	//		: : :
	// L	L 0 0
	// :	0 1 0
	//		1 1 0
	//		: : :
	//		L 1 0
	//		: : :
	//		L L 0
	//		1 0 1
	//		L L L
	// size of data: LxLxL
	void AddFaceData(int fID, float *data);

	// should be called if using GPU
	void LoadCudaData();
	void ClearCudaData();

	// Optimization
	int optimize(Options &options);

	// Solution
	int getSolution(int vID)
	{
		return (vID >= 0 && vID < _V) ? v_array[vID].min_label() : 0;
	}

	// Energy 
	float getEnergy(int *sol)
	{
		float sum = 0;
		// singleton terms
		for (int i = 0; i < _V; i ++)
			sum += v_array[i].theta_single[sol[i]];
		int l0, l1, l2;
		for (int f = 0; f < _F; f ++) {
			l0 = sol[f_array[f].vID[0]];
			l1 = sol[f_array[f].vID[1]];
			l2 = sol[f_array[f].vID[2]];
			sum += f_array[f].tri_energy(l0, l1, l2);
		}
		return sum;
	}



/////////////////////////////////////////////////////////////
////////////////////// Implementation ///////////////////////
/////////////////////////////////////////////////////////////

private: 

#define L_0(n)	(n % _L) // 
#define L_1(n)	((n / _L) % _L)
#define L_2(n)	((n / (_L*_L)) % _L)

	struct vertex
	{
		float *theta_single; // size : n_L 
		int _L;

		vertex (int L, float *addr) : _L(L), theta_single(addr) {}
		vertex& operator= (const vertex &v) { _L = v._L; theta_single = v.theta_single; return *this; }

		int min_label ();
		float single_energy (int l) { return (l >= 0 && l < _L) ? theta_single[l] : -1; }
	};
	struct face
	{
		float *theta_tri; // size : n_L^3
		int _L;
		int vID[3]; // index to the vertex

		face (int L, float *addr) : _L(L), theta_tri(addr) {}
		face& operator= (const face &f) { _L = f._L; theta_tri = f.theta_tri; return *this; }

		// v: vertex idx {0, 1, 2}, l: label
		float min_jk (int v, int l)
		{
			float min_val = 1e10;
			switch (v) {
			case 0:
				for (int i = l; i < _L*_L*_L; i += _L) {
					if (min_val > theta_tri[i])
						min_val = theta_tri[i];
				}
				break;
			case 1:
				for (int i = l*_L; i < _L*_L*_L; i += _L*_L)
					for (int j = 0; j < _L; j++) {
						if (min_val > theta_tri[i])
							min_val = theta_tri[i];
					}
				break;
			case 2:
				for (int i = l*_L*_L; i < (l+1)*_L*_L; i ++) {
					if (min_val > theta_tri[i])
						min_val = theta_tri[i];
				}
				break;
			}
			return min_val;
		}

		// add message w_diff to all (uw,ij) \theta_{uvw;lij}...
		void add_msg (int v, int l, float dif)
		{
			switch (v) {
			case 0:
				for (int i = l; i < _L*_L*_L; i += _L)
						theta_tri[i] += dif;
				break;
			case 1:
				for (int i = l*_L; i < _L*_L*_L; i += _L*_L)
					for (int j = 0; j < _L; j++) 			
						theta_tri[i+j] += dif;
				break;
			case 2:
				for (int i = l*_L*_L; i < (l+1)*_L*_L; i ++)	
						theta_tri[i] += dif;
				break;
			}
		}

		float tri_energy (int l0, int l1, int l2) { return theta_tri[l0 + l1 * _L + l2 * _L * _L]; }	
	};

	std::vector<vertex> v_array;
	std::vector<face> f_array;
	
	const int _V, _F, _L; // number of vertices, faces and labels

	float *arrayVert; // memory for singleton energy, size : _V * _L
	float *arrayFace; // memory for triplet energy, size : _F * _L * _L * _L
	int *fIdxArray; // memory for face to vertex index, size : _F * 3

	std::vector< std::vector<int> > ff_list; // for GPU implementation

	void randPermutation (int n, std::vector<int> &p_array);// permute [1...n] randomly
public:
	// Maximal independent faces
	void maximalIndFaces ();

	// Optimization techniques
	int optimize_CPU_FixedOrder (Options &options);
	int optimize_CPU_RandOrder (Options &options);
	int optimize_CPU_RandPerFace (Options &options);
	int optimize_GPU (Options &options);
	
};




#endif // _TRIMRFENERGY_H_